

import SwiftUI

struct EmojiMemoryGameView: View {
    @ObservedObject var game:EmojiMemoryGame //定义一个属性包装器，由启动入口指定使用的VM
    @Namespace private var dealingNamespace//定义一个命名空间
    var body: some View{
        //使用ZStack 让deckBody底部对齐
        ZStack(alignment: .bottom){
            VStack{
                gameBody
                HStack{
                    restart
                    Spacer()
                    shuffle
                }
                .padding(.horizontal)
            }
            deckBody//发牌位于 重新开始 和 重新洗牌 之间
        }.padding()
    }
    @State private var dealt = Set<Int>()//定义一个Int类型的空集合
    
    private func deal(_ card: EmojiMemoryGame.Card) {
        dealt.insert(card.id)//将卡片id插入到集合里
    }
    //返回卡片在集合里存在的状态，在 = false，不在 = true
    private func isUndealt(_ card: EmojiMemoryGame.Card)  -> Bool{
        !dealt.contains(card.id)//contains检查当前卡片是否存在于集合中
    }
    //延时发牌动画
    private func dealAnimation(for card:EmojiMemoryGame.Card) -> Animation{
        var delay = 0.0 //初始化延时Double
        if let index = game.cards.firstIndex(where: {$0.id == card.id}){
            //算出当前卡片需要延时多少
            delay = Double(index) * (CardConstants.totalDealDuration / Double(game.cards.count))
        }
        //返回动画执行时间0.5,并延时delay秒
        return Animation.easeInOut(duration: CardConstants.dealDuration).delay(delay)
    }
    //计算出当前卡片的合适zIndex值
    private func zIndex(for card:EmojiMemoryGame.Card) -> Double{
        -Double(game.cards.firstIndex(where: {$0.id == card.id}) ?? 0)
    }
    var gameBody: some View{
        AspectVGrid(items:game.cards,aspctRatio:CardConstants.aspectRatio){card in//新增一个视图组合器
            if isUndealt(card) || (card.isMatched && !card.isFaceUp){
                Color.clear//将原来的Rectangle()使用Color替换
            }else{
                CardView(card: card)
                    .matchedGeometryEffect(id: card.id, in: dealingNamespace)
                    .padding(4)
                    .transition(AnyTransition.asymmetric(insertion: .identity, removal: .scale))
                    .zIndex(zIndex(for:card))
                    .onTapGesture {
                        game.choose(card)
                    }
            }
        }
        .foregroundColor(CardConstants.color)
        .padding(.horizontal)
    }
    //建立一个洗好牌的待发的视图
    var deckBody:some View{
        ZStack{
            //通过filter过滤掉哪些还未处理的卡片,原型：filter{isUndealt($0)}
            ForEach(game.cards.filter(isUndealt)){ card in
                CardView(card: card)//通过ZStack让牌叠起来
                    .zIndex(zIndex(for:card))
                    .matchedGeometryEffect(id: card.id, in: dealingNamespace)
                    //设置一个与已发牌相关的过渡，插入时渐隐，删除时缩小
                    .transition(AnyTransition.asymmetric(insertion: .opacity, removal: .identity))
            }
        }
        .onTapGesture {
            for card in game.cards{
                withAnimation(dealAnimation(for: card)){ //通过当前card计算出需要延时多久
                    deal(card)//通过延时做为计时器，让卡片慢慢进入集合
                }
            }
        }
        //限制视图显示大小为固定值
        .frame(width: CardConstants.undealtWidth, height: CardConstants.undealtHeight)
        .foregroundColor(CardConstants.color)
    }
    //本视图里需要控制的一些参数
    private struct CardConstants{
        static let color = Color.red //卡片颜色
        static let aspectRatio:CGFloat = 2/3
        static let dealDuration: Double = 0.5
        static let totalDealDuration:Double = 2
        static let undealtHeight:CGFloat = 90//未发牌的高度
        static let undealtWidth = undealtHeight * aspectRatio//未发牌的宽度
    }
    
    //一个洗牌的按钮
    var shuffle:some View{
        Button("重新洗牌"){
            withAnimation{
                game.shuffle()//通过ViewModel调用Model里的洗牌功能
            }
        }
    }
    var restart:some View{
        Button("重新开始"){
            withAnimation{
                dealt = [] //初始化dealt集体
                game.restart()//通过ViewModel调用Model里的洗牌功能
            }
        }
    }
    
}

struct CardView:View {
    let card:MemoryGame<String>.Card
    @State private var animatedBonusRemaining:Double = 0//初始化为0
    var body: some View{
        GeometryReader{ geomery in
            ZStack{
                Group{
                    if card.isConsumingBonusTime{
                        Pie(startAngle: Angle(degrees: 0-90), endAngle: Angle(degrees: (1-animatedBonusRemaining)*360-90))
                            .onAppear{
                                animatedBonusRemaining = card.bonusRemaining//剩余奖金时间的百分比
                                //利用duration时间切片的属性，实现倒计时机制(还有多少时间奖金就到期了)
                                withAnimation(.linear(duration: card.bonusTimeRemaining)){
                                    animatedBonusRemaining = 0
                                }
                            }
                    }else{
                        Pie(startAngle: Angle(degrees: 0-90), endAngle: Angle(degrees: (1-card.bonusRemaining)*360-90))
                    }
                }
                .padding(5).opacity(0.5)
                Text(card.content)
                    //匹配后让其旋转360度
                    .rotationEffect(Angle.degrees(card.isMatched ? 360 : 0))
                    //repeatForever使其不断重复执行动画效果,autoreverses是否反向运动
                    .animation(Animation.easeInOut(duration: 2).repeatForever(autoreverses: false))
                    .font(Font.system(size: DrawingConstants.fontSize))
                    .scaleEffect(scale(thatFits:geomery.size))
            }
            //.animation(.easeInOut)//此优先权更高会导致withAnimation计时失败
            //引用自己定义的Cardify并传入判断参数
            .cardify(isFaceUp: card.isFaceUp)
        }
    }
    //计算出当前字体需要缩放的比例
    private func scale(thatFits size:CGSize) -> CGFloat{
        min(size.width,size.height) / (DrawingConstants.fontSize / DrawingConstants.fontScale)
    }
    //参数控制器移出Cardify部分
    private struct DrawingConstants{
        static let fontScale: CGFloat = 0.7//Emoji缩放比例
        static let fontSize: CGFloat = 32//固定Emoji的字体大小
    }
}

struct ContentView_Previews: PreviewProvider {
    static let game = EmojiMemoryGame()//定义为static，全局可用
    static var previews: some View {
        game.choose(game.cards.first!)//直接调用game
        return EmojiMemoryGameView(game: game)
    }
}
