import Foundation //此包导入了Array、String、Int等基本的数据类型
//对CardContent提出新的要求符合Equatable协议
struct MemoryGame<CardContent> where CardContent: Equatable{
    private(set) var cards:Array<Card>//卡片只读,在任何情况下只能使用内部函数对数据进行CRUD
    private var indexOfTheOneAndOnlyFaceUpCard:Int?{
        get{ cards.indices.filter{ cards[$0].isFaceUp }.oneAndOnly }
        set{ cards.indices.forEach{ cards[$0].isFaceUp = ($0 == newValue) } }
    }
    
    //通过计算属性已将数据同步
    mutating func choose(_ card:Card) {
        if let chosenIndex = cards.firstIndex(where: { $0.id == card.id}) ,
           !cards[chosenIndex].isFaceUp,//增加要求被选择的卡片isFaceUp为false
           !cards[chosenIndex].isMatched//要求被选择的卡片isMatched为false
        {
            //indexOfTheOneAndOnlyFaceUpCard的get计算属性在这里使用
            if let potentialMatchIndex = indexOfTheOneAndOnlyFaceUpCard{
                if cards[chosenIndex].content == cards[potentialMatchIndex].content{
                    cards[chosenIndex].isMatched = true
                    cards[potentialMatchIndex].isMatched = true
                }
                //这条代码已被上面的计算属性所替代 indexOfTheOneAndOnlyFaceUpCard = nil
                cards[chosenIndex].isFaceUp = true//选中的卡片朝上
            } else {
                //set计算属性在这里生效
                indexOfTheOneAndOnlyFaceUpCard = chosenIndex
            }
        }
    }
    //洗牌，让卡片内容随机排列
    mutating func shuffle() {
        cards.shuffle()
    }
    //初始化卡片个数及使用闭包确定卡片的内容是什么
    init(numberOfPairsOfCards: Int, createCardContent:(Int) -> CardContent) {
        cards = []//初始化卡片容器
        //让卡片在容器里成对出现，通过numberOfPairsOfCards指定个数
        for pairIndex in 0..<numberOfPairsOfCards{
            let content = createCardContent(pairIndex)//使用闭包来获取到卡片的内容（传入的是当前索引Int实参）
            cards.append(Card(content: content, id: pairIndex * 2))//容器增加一个卡片内容
            cards.append(Card(content: content, id: pairIndex * 2 + 1))//容易再增加一个内容相同的卡片
        }
        cards.shuffle()//洗牌
    }
    
//Card被算法重写了
struct Card: Identifiable {
    //isFaceUp增加了属性观察者，当翻开后开始计时，背对暂停计时
    var isFaceUp = false {
        didSet {
            if isFaceUp {
                startUsingBonusTime()
            } else {
                stopUsingBonusTime()
            }
        }
    }
    //增加属性观察者，一但匹配停止计时
    var isMatched = false {
        didSet {
            stopUsingBonusTime()
        }
    }
    let content: CardContent
    let id: Int
    
    // MARK: - 计时器算法
    
    //额外的时间限制，超时后奖励为0
    var bonusTimeLimit: TimeInterval = 6
    
    // 这张牌朝上多久了
    private var faceUpTime: TimeInterval {
        if let lastFaceUpDate = self.lastFaceUpDate {
            return pastFaceUpTime + Date().timeIntervalSince(lastFaceUpDate)
        } else {
            return pastFaceUpTime
        }
    }
    // 上次这张牌是面朝上的(现在仍然是面朝上)
    var lastFaceUpDate: Date?
    //这张卡过去的累计时间 (例如，如果当前是正面，则不包含当前时间)
    var pastFaceUpTime: TimeInterval = 0
    
    // 还有多少时间奖金就到期了
    var bonusTimeRemaining: TimeInterval {
        max(0, bonusTimeLimit - faceUpTime)
    }
    // 剩余奖金时间的百分比
    var bonusRemaining: Double {
        (bonusTimeLimit > 0 && bonusTimeRemaining > 0) ? bonusTimeRemaining/bonusTimeLimit : 0
    }
    // 是否在奖金期间匹配该卡
    var hasEarnedBonus: Bool {
        isMatched && bonusTimeRemaining > 0
    }
    // 无论我们目前是否正面向上，还是无可匹敌的，奖金窗口还没有用完
    var isConsumingBonusTime: Bool {
        isFaceUp && !isMatched && bonusTimeRemaining > 0
    }
    
    // 当卡片转换为正面状态时调用开始计时
    private mutating func startUsingBonusTime() {
        if isConsumingBonusTime, lastFaceUpDate == nil {
            lastFaceUpDate = Date()
        }
    }
    // 当牌面朝下(或匹配)时调用 暂时/停止计时
    private mutating func stopUsingBonusTime() {
        pastFaceUpTime = faceUpTime
        self.lastFaceUpDate = nil
    }
}
    
}

extension Array where Element:Hashable{
    var oneAndOnly:Element?{ (count == 1) ? first : nil }
}
